/*!\file animeManager.c
 *
 * \brief Bibliothèque de gestion des écrans d'animation et de mises à
 * jour en fonction du son.
 * \author Farès BELHADJ, amsi@ai.univ-paris8.fr
 * \date April 29 2014
 */
#include "animeManager.h"
#include <assert.h>
#include <stdlib.h>

static int  in(void *func, void **funcList, int n);
static void add(void *func, void ***funcListp, int *n, int *s);
static void callAllWithState(anime_t * animations, int state);
static void drawOrUpdateWithAudio(anime_t * animations, int state);

static Uint32 _startingTime = 0;

/*!\brief Demande l'état initialisation à tous les éléments du tableau d'animations.
 * \param animations tableau d'animations se terminant par un élément dont le champ first est à NULL.
 */
void amInit(anime_t * animations) {
  callAllWithState(animations, S_INIT);
}

/*!\brief Demande l'état libération à tous les éléments du tableau d'animations.
 * \param animations tableau d'animations se terminant par un élément dont le champ first est à NULL.
 */
void amFree(anime_t * animations) {
  callAllWithState(animations, S_FREE);
}

/*!\brief donne le top-départ de la démo. Cette fonction doit être
 * appelée juste avant le début de la boucle infinie d'affichage (soit
 * juste avant le premier S_DRAW obtenu en appelant amDraw).
 */
void amStartingSignal(void) {
  _startingTime = SDL_GetTicks();
}

/*!\brief renvoie le nombre de milisecondes passée depuis le
 * top-départ donné par \a amStartingSignal.
 */
Uint32 amGetTicks(void) {
  return SDL_GetTicks() - _startingTime;
}

/*!\brief joue parmi les animations passées en argument l'animation en
 * cours (devant être jouée au temps t).
 *
 * \param animations tableau d'animations se terminant par un élément dont le champ first est à NULL.
 */
void amDraw(anime_t * animations) {
  drawOrUpdateWithAudio(animations, S_DRAW);
}

/*!\brief met à jour parmi les animations passées en argument l'animation en
 * cours.
 *
 * \param animations tableau d'animations se terminant par un élément dont le champ first est à NULL.
 */
void amUpdateWithAudio(anime_t * animations) {
  drawOrUpdateWithAudio(animations, S_UPDATE_WITH_AUDIO);
}

/*!\brief regarde si le pointeur \a func est un élément du tableau
 * \a funcList 
 * \param func un élément dont il faut tester la présence.
 * \param funcList un tableau d'éléments.
 * \param n le nombre d'éléments dans \a funcList.
 * \return vrai si l'élément est présent, faux sinon.
 */
static int in(void *func, void **funcList, int n) {
  int i;
  for(i = 0; i < n; i++)
    if(func == funcList[i])
      return 1;
  return 0;
}

/*!\brief ajoute le pointeur \a func dans le tableau d'éléments 
 * *\a funcListp 
 * \param func un élément dont il faut tester la présence.
 * \param funcListp le pointeur vers le tableau d'éléments.
 * \param n le pointeur vers le nombre d'éléments dans *\a funcListp.
 * \param s le pointeur vers la taille, en nombre d'éléments, de *\a funcListp.
 */
static void add(void *func, void ***funcListp, int *n, int *s) {
  if(*s == 0) {
    *n = 0;
    *funcListp = calloc((*s = 16), sizeof **funcListp);
    assert(*funcListp);
  } else if(*n == *s) {
    *funcListp = realloc(*funcListp, (*s *= 2) * sizeof **funcListp);
    assert(*funcListp);
  }
  (*funcListp)[(*n)++] = func;
} 

/*!\brief Parcourt un tableau d'animations (possiblement dupliquées)
 * et les appelle une et une seule fois avec l'état \a state.
 * \param animations tableau d'animations se terminant par un élément dont le champ first est à NULL.
 * \param state état à utiliser pour chaque appel.
 */
static void callAllWithState(anime_t * animations, int state) {
  int n = 0, s = 0;
  void **called = NULL;
  assert(animations);
  while(animations->first) {
    if(!in(animations->first, called, n)) {
      animations->first(state);
      add(animations->first, &called, &n, &s);
    }
    if(animations->last && !in(animations->last, called, n)) {
      animations->last(state);
      add(animations->last, &called, &n, &s);
    }
    if(animations->transition && !in(animations->transition, called, n)) {
      animations->transition(NULL, NULL, 0, 0, state);
      add(animations->transition, &called, &n, &s);
    }
    animations++;
  }
  if(called)
    free(called);
}

/*!\brief joue ou met à jour parmi les animations passées en argument
 * l'animation en cours (devant être jouée/mise à jour au temps t).
 *
 * \param animations tableau d'animations se terminant par un élément
 * dont le champ first est à NULL.
 * \param state état devant être appliqué : jouer (S_DRAW) ou mettre à
 * jour (S_UPDATE_WITH_AUDIO)
 */
static void drawOrUpdateWithAudio(anime_t * animations, int state) {
  int i;
  static int cur = 0;
  Uint32 t = 0, at = 0;
  if(!animations[cur].first)
    goto noMoreAnimationsFlag;
  for(i = 0; i < cur; i++) at += animations[i].time;
  t = amGetTicks() - at;
  if(t >= animations[cur].time) {
    t -= animations[cur].time;
    cur = (cur + 1);
    if(!animations[cur].first)
      goto noMoreAnimationsFlag;
  }
  if(animations[cur].transition)
    animations[cur].transition(animations[cur].first, animations[cur].last, animations[cur].time, t, state);
  else
    animations[cur].first(state);
  return;
 noMoreAnimationsFlag:
  if(state == S_DRAW) 
    exit(0); // mettre cur = 0; pour tourner en boucle
}

