/*!\file miscHelper.c
 *
 * \brief Divers éléments pouvant être utiles pour la démo.
 * \author Farès BELHADJ, amsi@ai.univ-paris8.fr
 * \date May 5, 2014
 */

#include "miscHelper.h"

/*!\brief identifiant du vertex array object */
static GLuint _vao = 0;
/*!\brief identifiant du buffer de data */
static GLuint _buffer = 0;
/*!\brief comptele nombre de mhInit appelés, utile pour savoir si
 * exécution effective de mhQuit */
static int _hasInit = 0;
/*!\brief pointeur vers le flux audio.
 * \see mhGetAudioStream
 * \see mhSetAudioStream
 */
static Uint8 * _audioStream = NULL;
/*!\brief longueur du flux audio.
 * \see mhGetAudioStreamLength
 * \see mhSetAudioStream
 */
static int _audioStreamLength = 0;


/*!\brief Initialisation des éléments de cette bibliothèque. 
 *
 * \a _vao est créé et fait référence à un quad mappé de triangles
 * strips représentant l'écran.
*/
void mhInit(void) {
  GLfloat data[] = {
    /* 4 coordonnées de sommets */
    -1.f, -1.f, 1.f, -1.f,
    -1.f,  1.f, 1.f,  1.f,
    /* 2 coordonnées de texture par sommet */
    0.0f, 0.0f, 1.0f, 0.0f, 
    0.0f, 1.0f, 1.0f, 1.0f
  };
  if(_hasInit++) return;
  glGenVertexArrays(1, &_vao);
  glBindVertexArray(_vao);
  glGenBuffers(1, &_buffer);
  glBindBuffer(GL_ARRAY_BUFFER, _buffer);
  glBufferData(GL_ARRAY_BUFFER, sizeof data, data, GL_STATIC_DRAW);
  glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, 0, (const void *)0);  
  glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 0, (const void *)((4 * 2) * sizeof *data));
  glBindBuffer(GL_ARRAY_BUFFER, 0);
  glBindVertexArray(0);
}

/*!\brief Libération des éléments de cette bibliothèque. 
*/
void mhQuit(void) {
  if(_hasInit <= 0) { _hasInit = 0; return; }
  if(!--_hasInit) return;
  if(_vao) {
    glDeleteVertexArrays(1, &_vao);
    _vao = 0;
  }
  if(_buffer) {
    glDeleteBuffers(1, &_buffer);
    _buffer = 0;
  }
}

/*!\brief retourne l'identifiant du VAO représentant le quad mappé. 
 * \return l'identifiant du VAO représentant le quad mappé. 
 */
GLuint mhGetMappedQuadVAO(void) {
  return _vao;
}

/*!\brief renvoie le pointeur vers le flux audio.
 * \return le pointeur vers le flux audio.
 */
Uint8 * mhGetAudioStream(void) {
  return _audioStream;
}

/*!\brief renvoie la longueur du flux audio.
 * \return la longueur du flux audio.
 */
int mhGetAudioStreamLength(void) {
  return _audioStreamLength;
}

/*!\brief met à jour le pointeur vers le flux audio et sa longueur.
 */
void mhSetAudioStream(Uint8 * audioStream, int audioStreamLength) {
  _audioStream = audioStream;
  _audioStreamLength = audioStreamLength;
}

