/*!\file window.c
 *
 * \brief Utilisation de la SDL2 et d'OpenGL 3+ et de la bibliothèque
 * animeManager afin de construire une démo.
 *
 * Ici est géré l'ouverture de la fenêtre SDL2/OpenGL ainsi que
 * l'ordonnancement des animations. Apriori les seuls éléments à
 * modifier ici lors de votre intégration sont le tableau static \ref
 * _animations et la fonction qui charge l'audio \ref initAudio.
 *
 * \author Farès BELHADJ, amsi@ai.univ-paris8.fr
 * \date May 05, 2014
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <time.h>

#include <gl4du.h>
#include <animeManager.h>
#include <animations.h>
#include <miscHelper.h>
#include <SDL_mixer.h>

/* Prototypes des fonctions statiques contenues dans ce fichier C. */
static SDL_Window * initWindow(int w, int h, SDL_GLContext * poglContext);
static void init(SDL_Window * win);
static void initAudio(void);
static void quit(void);
static void resizeGL(SDL_Window * win);
static void loop(SDL_Window * win);
static void manageEvents(SDL_Window * win);
static void draw(void);

/*!\brief tableau contenant les animations sous la forme de timeline,
 * ce tableau se termine toujours par l'élémént {0, NULL, NULL,
 * NULL} */
static anime_t _animations[] = {
  { 1000, exemple_d_animation_06, NULL, NULL },
  { 2800, exemple_d_animation_06, exemple_d_animation_02, exemple_de_transition_01 },
  /*{ 1000, exemple_d_animation_01, NULL, NULL },
    { 2800, exemple_d_animation_01, exemple_d_animation_02, exemple_de_transition_01 },*/
  { 2000, exemple_d_animation_02, NULL, NULL },
  { 1000, exemple_d_animation_02, exemple_d_animation_05, exemple_de_transition_01 },
  { 5000, exemple_d_animation_05, NULL, NULL },
  { 3000, exemple_d_animation_00, NULL, NULL },
  { 2800, exemple_d_animation_00, exemple_d_animation_03, exemple_de_transition_01 },
  {12000, exemple_d_animation_03, NULL, NULL },
  { 2800, exemple_d_animation_03, exemple_d_animation_04, exemple_de_transition_01 },
  {12000, exemple_d_animation_04, NULL, NULL },
  {    0, NULL, NULL, NULL } /* Toujours laisser à la fin */
};

/*!\brief dimensions initiales de la fenêtre */
static GLfloat _dim[] = {1024, 768};
/*!\brief pointeur vers la (future) fenêtre SDL */
static SDL_Window * _win = NULL;
/*!\brief pointeur vers le (futur) contexte OpenGL */
static SDL_GLContext _oglContext = NULL;
/*!\brief identifiant du framebuffer utilisé */
static GLuint _fbo = 0;
/*!\brief identifiant de la texture liée en écriture au framebuffer (la texture dans laquelle nous allons dessiner) */
static GLuint _wTexId = 0;
/*!\brief pointeur vers la musique chargée par SDL_Mixer */
static Mix_Music * _mmusic = NULL;

/*!\brief La fonction principale initialise la bibliothèque SDL2
 * (vidéo et son), demande la création de la fenêtre SDL et du
 * contexte OpenGL par l'appel à \ref initWindow, initialise OpenGL et
 * autres avec la fonction \ref init et lance la boucle (infinie)
 * principale.
 */
int main(int argc, char ** argv) {
  if(SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO) < 0) {
    fprintf(stderr, "Erreur lors de l'initialisation de SDL :  %s", SDL_GetError());
    return -1;
  }
  atexit(SDL_Quit);
  if((_win = initWindow(_dim[0], _dim[1], &_oglContext))) {
    init(_win);
    initAudio();
    loop(_win);
  }
  return 0;
}

/*!\brief Cette fonction créé la fenêtre SDL de largeur \a w et de
 *  hauteur \a h, le contexte OpenGL \a poglContext et stocke le
 *  pointeur dans poglContext. Elle retourne le pointeur vers la
 *  fenêtre SDL.
 *
 * Le contexte OpenGL créé est en version 3 pour
 * SDL_GL_CONTEXT_MAJOR_VERSION, en version 2 pour
 * SDL_GL_CONTEXT_MINOR_VERSION et en SDL_GL_CONTEXT_PROFILE_CORE
 * concernant le profile. Le double buffer est activé et le buffer de
 * profondeur est en 24 bits.
 *
 * \param w la largeur de la fenêtre à créer.
 * \param h la hauteur de la fenêtre à créer.
 * \param poglContext le pointeur vers la case où sera référencé le
 * contexte OpenGL créé.
 * \return le pointeur vers la fenêtre SDL si tout se passe comme
 * prévu, NULL sinon.
 */
static SDL_Window * initWindow(int w, int h, SDL_GLContext * poglContext) {
  SDL_Window * win = NULL;
  SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 3);
  SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 2);
  SDL_GL_SetAttribute(SDL_GL_CONTEXT_PROFILE_MASK, SDL_GL_CONTEXT_PROFILE_CORE);
  SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
  SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);
  if( (win = SDL_CreateWindow("DemoHelper", SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED, 
			      w, h, SDL_WINDOW_OPENGL | SDL_WINDOW_FULLSCREEN_DESKTOP | SDL_WINDOW_SHOWN)) == NULL ) {
    fprintf(stderr, "Problème lors de l'ouverture de la fenêtre : %s\n", SDL_GetError());
    return NULL;
  }
  if( (*poglContext = SDL_GL_CreateContext(win)) == NULL ) {
    fprintf(stderr, "Problème lors de la création du context GL : %s\n", SDL_GetError());
    SDL_DestroyWindow(win);
    return NULL;
  }
  fprintf(stderr, "Version d'OpenGL : %s\n", glGetString(GL_VERSION));
  fprintf(stderr, "Version de shaders supportes : %s\n", glGetString(GL_SHADING_LANGUAGE_VERSION));  
  atexit(quit);
  return win;
}

/*!\brief Cette fonction initialise les paramètres et éléments OpenGL
 * ainsi que les divers données et fonctionnalités liées à la gestion
 * des animations.
 *
 * \param win le pointeur vers la fenêtre SDL pour laquelle nous avons
 * attaché le contexte OpenGL.
 */
static void init(SDL_Window * win) {
  glClearColor(0.2f, 0.2f, 0.2f, 0.0f);
  resizeGL(win);
  mhInit();
  glGenFramebuffers(1, &_fbo);
  glGenTextures(1, &_wTexId);
  glBindTexture(GL_TEXTURE_2D, _wTexId);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, _dim[0], _dim[1], 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
  glBindTexture(GL_TEXTURE_2D, 0);

  glBindFramebuffer(GL_FRAMEBUFFER, _fbo);
  glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D,  _wTexId,  0);
  amInit(_animations);
  glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

/*!\brief Cette fonction est appelée quand l'audio est joué et met 
 * dans \a stream les données audio de longueur \a len.
 * \param udata pour user data, données passées par l'utilisateur, ici NULL.
 * \param stream flux de données audio.
 * \param len longueur de \a stream.
 */
static void mixCallback(void *udata, Uint8 *stream, int len) {
  mhSetAudioStream(stream, len);
  amUpdateWithAudio(_animations);
  mhSetAudioStream(NULL, 0);
}

/*!\brief Cette fonction initialise les paramètres SDL_Mixer et charge le fichier audio.
 */
static void initAudio(void) {
  int mixFlags = MIX_INIT_MP3, res;
  res = Mix_Init(mixFlags);
  if( (res & mixFlags) != mixFlags ) {
    fprintf(stderr, "Mix_Init: Erreur lors de l'initialisation de la bibliotheque SDL_Mixer\n");
    fprintf(stderr, "Mix_Init: %s\n", Mix_GetError());
    exit(-3);
  }
  if(Mix_OpenAudio(22050, AUDIO_S16LSB, 2, 1024) < 0)
    exit(-4);
  if(!(_mmusic = Mix_LoadMUS("happy.ogg"))) {
    fprintf(stderr, "Erreur lors du Mix_LoadMUS: %s\n", Mix_GetError());
    exit(-5);
  }
  Mix_SetPostMix(mixCallback, NULL);
}

/*!\brief Cette fonction est appelée au moment de sortir du programme
 *  (atexit), elle libère la fenêtre SDL \ref _win et le contexte
 *  OpenGL \ref _oglContext.
 */
static void quit(void) {
  if(_mmusic) {
    if(Mix_PlayingMusic())
      Mix_HaltMusic();
    Mix_FreeMusic(_mmusic);
    _mmusic = NULL;
  }
  Mix_CloseAudio();
  Mix_Quit();
  mhQuit();
  amFree(_animations);
  if(_wTexId)
    glDeleteTextures(1, &_wTexId);
  if(_fbo)
    glDeleteFramebuffers(1, &_fbo);
  if(_oglContext)
    SDL_GL_DeleteContext(_oglContext);
  if(_win)
    SDL_DestroyWindow(_win);
  gl4duClean(GL4DU_ALL);
}

/*!\brief Cette fonction paramétrela vue (viewPort) OpenGL en fonction
 * des dimensions de la fenêtre SDL pointée par \a win.
 *
 * \param win le pointeur vers la fenêtre SDL pour laquelle nous avons
 * attaché le contexte OpenGL.
 */
static void resizeGL(SDL_Window * win) {
  int w, h;
  SDL_GetWindowSize(win, &w, &h);
  _dim[0] = w;
  _dim[1] = h;
  glViewport(0, 0, _dim[0], _dim[1]);
}

/*!\brief Boucle infinie principale : gère les évènements, dessine,
 * imprime le FPS et swap les buffers.
 *
 * \param win le pointeur vers la fenêtre SDL pour laquelle nous avons
 * attaché le contexte OpenGL.
 */
static void loop(SDL_Window * win) {
  SDL_GL_SetSwapInterval(1);
  if(!Mix_PlayingMusic())
    Mix_PlayMusic(_mmusic, 1);
  amStartingSignal();
  for(;;) {
    manageEvents(win);
    draw();
    SDL_GL_SwapWindow(win);
    gl4duUpdateShaders();
  }
}

/*!\brief Cette fonction permet de gérer les évènements clavier et
 * souris via la bibliothèque SDL et pour la fenêtre pointée par \a
 * win.
 *
 * \param win le pointeur vers la fenêtre SDL pour laquelle nous avons
 * attaché le contexte OpenGL.
 */
static void manageEvents(SDL_Window * win) {
  SDL_Event event;
  while(SDL_PollEvent(&event)) 
    switch (event.type) {
    case SDL_KEYDOWN:
      switch(event.key.keysym.sym) {
      case SDLK_ESCAPE:
      case 'q':
	exit(0);
      default: break;
      }
      break;
    case SDL_KEYUP:
      break;
    case SDL_WINDOWEVENT:
      if(event.window.windowID == SDL_GetWindowID(win)) {
	switch (event.window.event)  {
	case SDL_WINDOWEVENT_RESIZED:
	  resizeGL(win);
	  break;
	case SDL_WINDOWEVENT_CLOSE:
	  event.type = SDL_QUIT;
	  SDL_PushEvent(&event);
	  break;
	}
      }
      break;
    case SDL_QUIT:
      exit(0);
    }
}

static void draw(void) {
  glBindFramebuffer(GL_FRAMEBUFFER, _fbo);
  glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D,  _wTexId,  0);
  amDraw(_animations);

  glUseProgram(0);

  glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);
  glBlitFramebuffer(0, 0, _dim[0], _dim[1], 0, 0, _dim[0], _dim[1], GL_COLOR_BUFFER_BIT, GL_LINEAR);
  glBindFramebuffer(GL_FRAMEBUFFER, 0);
}
